==> wfde-3.01.10.tar.gz: wfde-3.01.10/cpplib/WarUrl.cpp <==
#include "StdAfx.h"

#ifndef WAR_ASSERT_H_INCLUDED
#   define WAR_ASSERT_H_INCLUDED
#   include <assert.h>
#endif

#ifndef WAR_MALLOC_H_INCLUDED
#   define WAR_MALLOC_H_INCLUDED
#   include <malloc.h>
#endif

#include "WarUrl.h"   // class implemented

#ifndef WAR_VECTOR_INCLUDED
#   include <vector>
#endif

#ifndef WAR_UTF8_H
#   include "WarUtf8.h"
#endif

using namespace std;


/////////////////////////////// PUBLIC ///////////////////////////////////////

WarUrl& operator << (WarUrl& path, const war_svrpath_t& from)
{
    path.Add(from);
    return path;
}

WarUrl& operator << (WarUrl& path, const WarPathE from)
{
    path.Add(from);
    return path;
}


const char WarUrl::msWarHexToCharMap[127] =
{
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 0
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 10
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 20
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 30
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01, // 40
    0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x00,0x00, // 50
    0x00,0x00,0x00,0x00,0x00,0x0a,0x0b,0x0c,0x0d,0x0e, // 60
    0x0f,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 70
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 80
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x0a,0x0b,0x0c, // 90
    0x0d,0x0e,0x0f,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 100
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 110
    0x00,0x00,0x00,0x00,0x00,0x00,0x00                 // 120
};

//============================= LIFECYCLE ====================================

//============================= OPERATORS ====================================

//============================= OPERATIONS ===================================


void WarUrl::Create(war_ccstr_t type_name,
                    war_ccstr_t UserName,
                    war_ccstr_t Password,
                    war_port_t Port,
                    war_ccstr_t Host,
                    war_ccstr_t UrlPath,
                    war_ccstr_t Command)
                    throw(WarException)
{
    WarUrlType url_type;
    url_type.Create(type_name);
    Create(url_type, 
        UserName,
        Password,
        Port,
        Host,
        UrlPath,
        Command);
}

void WarUrl::Create(const WarUrlType& Type,
                    war_ccstr_t UserName,
                    war_ccstr_t Password,
                    war_port_t Port,
                    war_ccstr_t Host,
                    war_ccstr_t UrlPath,
                    war_ccstr_t Command)
                    throw(WarException)
{
    mUrlType = Type;
    if (UserName && *UserName)
    {
        mUserName = UserName;
        mHaveUserName = true;
    }
    
    if (Password && *Password)
    {
        mPassword = Password;
        mHavePassword = true;
    }
    
    if (Port)
    {
        mPort = Port;
        mHavePort = true;
    }
    
    if (Host && *Host)
    {
        mHostName = Host;
        mHaveHostName = true;
    }
    
    if (UrlPath && *UrlPath)
    {
        mUrlPath = UrlPath;
        mHaveUrlPath = true;
    }
    
    if (Command && *Command)
    {
        mCommand = Command;
        mHaveCommand = true;
    }
    
    BuildUrl();
}

void WarUrl::Reset()
{
    mUrlType.Reset();
    mPort = 0;
    mUserName = "";
    mPassword = "";
    mHostName = "";
    mUrlPath = "";
    mUrl = "";
    mCommand = "";
    mFilePath.Reset();
    
    mHavePort = 0;
    mHaveUserName = 0;
    mHavePassword = 0;
    mHaveHostName = 0;
    mHaveUrlPath = 0;
    mHaveCommand = 0;
}

void WarUrl::Normalize()
{
    mFilePath.Normalize();
    mFilePath.RemovePaddingSlash();
    if (mFilePath.mString.find(mFilePath.GetSlash()) == mFilePath.mString.npos)
        mFilePath << WAR_SYSSLASH;
}

//============================= ACCESS     ===================================
//============================= INQUIRY    ===================================


/////////////////////////////// PROTECTED  ///////////////////////////////////

/////////////////////////////// PRIVATE    ///////////////////////////////////


void WarUrl::ParseString(war_ccstr_t src)
{
    Reset();
    
    size_t url_len = 0;
    war_ccstr_t p = src;
    size_t len = strlen(src) + 8;
    vector<char> tmp_buffer(len);
    war_cstr_t buffer = (war_cstr_t)&tmp_buffer.front();
    war_cstr_t pp = buffer;
    war_ccstr_t pseg1 = NULL, pseg2 = NULL;
    
    try
    {
        mUrlType.Create(src);
        url_len = mUrlType.GetLenght(); 
    }
    catch(WarException& ex)
    {
        // Assume file if no type is set
#ifdef WIN32
        if (ex == WAR_ERR_INVALID_URL)
        {
            if (!(src 
                && isalpha(*src)
                && (src[1] == ':')
                && ((src[2] == '\\') || (src[2] == '/'))
                && ((src[3] != '\\') && (src[3] != '/'))))
            throw ex;    
        }
        else
#endif
        if (ex != WAR_ERR_NOT_AN_URL)
            throw ex;
        
        mUrlType.Create("file:///");
        goto skip_hostname;
    }
    
    p += url_len;
    
    if (mUrlType.IsHostscanRequiered() && (*p != '/'))
    {
        // We have host, or username, or password, or something...
        // Extract the segment...
        pseg1 = pp = buffer;
        bool bHaveAlpha = false, bHaveCol = false;
        
        while(*p && (*p != '/'))
        {
            switch(*p)
            {
            case ':':
                ++p;
                if (bHaveCol)
                    WarThrow(WarError(WAR_ERR_INVALID_URL), NULL);
                bHaveCol = true;
                *pp++ = 0;
                pseg2 = pp;
                break;
            case '@':
                ++p;
                if (bHaveAlpha)
                    WarThrow(WarError(WAR_ERR_INVALID_URL), NULL);
                bHaveAlpha = true;
                *pp++ = 0;
                mHaveUserName = true;
                if (*pseg1)
                    Decode(mUserName, pseg1);
                if (bHaveCol)
                {
                    // second segment is password
                    mHavePassword = true;
                    assert(pseg2 != NULL);
                    if (*pseg2)
                        Decode(mPassword.mString, pseg2);
                }
                // Reset
                bHaveCol = false;
                pseg1 = pp;
                pseg2 = NULL;
                break;
            default:
                *pp++ = *p++;
                break;
            }
        }
        *pp = 0;
        if (*pseg1)
        {
            mHaveHostName = true;
            Decode(mHostName, pseg1);
        }
        if (bHaveCol)
        {
            mHavePort = true;
            if (!*pseg2)
                WarThrow(WarError(WAR_ERR_INVALID_URL), NULL);
            mPort = atoi(pseg2);
        }
    }

skip_hostname:
    
    if (mUrlType.IsHostscanRequiered() && (*p == '/'))
        ++p;
    
#ifdef WIN32
    else if (!url_len
        && (p[0] == '\\')
        && (p[1] == '\\')
        && (p[2] != '\\')
        && (p[2] != 0))
    {
        // UNC path. Convert to URL
        pseg1 = pp;
        p += 2;
        while(!WarSysSlash::IsSlash(*p))
            *pp++ = *p++;
        
        mHaveHostName = true;
        if (*pseg1)
            mHostName = pseg1;
        
        if (*p)
            ++p;
    }
#endif
    
    // Until now, we have decoded the URL type, and the user@host part.
    // We now have to decode the url-path and command
    
    if (*p)
    {
        
        pseg1 = pp;
        pseg2 = NULL;

        // Parse to the command mark
        while(*p && !mUrlType.IsCmdDelimiter(*p))
            *pp++ = *p++;

        *pp++ = 0;
        pseg2 = pp;
        if (*p)
            ++p;
        while(*p)
            *pp++ = *p++;
        *pp = 0;
        
        if (*pseg1 || pseg2)
        {
            assert((pp - buffer) - len);
            mHaveUrlPath = true;
            if (*pseg1)
                Decode(mUrlPath, pseg1);
        }
        
        if (pseg2 && *pseg2)
        {
            mHaveCommand = true;
            if (*pseg2)
                Decode(mCommand, pseg2);
        }
        
    }
    
    BuildUrl();
}

void WarUrl::BuildUrl()
{
    string TmpBuf;
    
    mUrl = GetTypeName();
    
    if (mHaveUserName)
    {
        if (!mUserName.empty())
            mUrl << Encode(TmpBuf, mUserName.c_str());
    }
    
    if (mHavePassword)
    {
        mUrl << ':';
        if (!mPassword.GetValue().empty())
            mUrl << Encode(TmpBuf, mPassword.GetValue().c_str());
    }
    
    if (mHaveUserName || mHavePassword)
        mUrl << '@';
    
    if (mHaveHostName)
    {
        if (!mHostName.empty())
            mUrl << Encode(TmpBuf, mHostName.c_str());
    }
    
    if (mHavePort)
    {
        mUrl << ':';
        WarItoa(mUrl.mString, (unsigned)mPort);
    }
    
    mUrl << '/';
    
    if (mHaveUrlPath)
    {
        if (!mUrlPath.empty())
            mUrl << Encode(TmpBuf, mUrlPath.c_str(), ENC_PATH);
    }
    
    if (mHaveCommand)
    {
        char url_delimiter = mUrlType.GetCmdDelimiter();
        if (!url_delimiter)
            WarThrow(WarError(WAR_ERR_INVALID_ARGUMENT), 
                "URL command not available");
        mUrl << url_delimiter;
        if (!mCommand.empty())
            mUrl << Encode(TmpBuf, mCommand.c_str(), ENC_PATH);
    }
    
    if (mUrlType == "file")
    {
        // Make a system file path!
        mFilePath.Reset();
#ifdef WIN32
        if (!mHostName.empty())
        {
            mFilePath << WAR_DSYSSLASH
                << WarUtf8(mHostName).GetUnicode()
                << WAR_SYSSLASH;
        }
#endif
        if (mUrlPath.empty())
            mFilePath << WAR_CDIR;
        else
            mFilePath << WarUtf8(mUrlPath).GetUnicode();
        mFilePath.Normalize();
    }

    if (!mHavePassword || GetPassword().empty())
    {
        mSafeUrl = mUrl.GetValue();
    }
    else
    {
        // Hide the password
        war_ccstr_t p = mUrl.GetValue().c_str();
        char *buf = (char *)alloca(mUrl.GetValue().size() + 8);
        char *pp = buf;

        while(*p && (*p != ':'))
            *pp++ = *p++;
        if (*p == ':')
            *pp++ = *p++;
        if (*p == '/')
            *pp++ = *p++;
        if (*p == '/')
            *pp++ = *p++;
        while(*p && (*p != ':'))
            *pp++ = *p++;
        if (*p == ':')
            *pp++ = *p++;
        while(*p && (*p != '@'))
            +p++;
        *pp++ = '*';
        *pp++ = '*';
        *pp++ = '*';
        *pp++ = '*';
        while(*p)
            *pp++ = *p++;
        *pp = 0;
        mSafeUrl = buf;
    }
}

std::string& WarUrl::Encode(std::string& Dst, 
                            war_ccstr_t Src, 
                            const EncodeModeE encMode)
{
    war_ccstr_t p = Src;
    size_t max_len = (strlen(Src) * 3)+ 1;
    
    war_cstr_t buffer = (war_cstr_t)alloca(max_len);
    war_cstr_t pp = buffer;
    
    while(*p)
    {
        if (*p == '/')
        {
            if ((encMode == ENC_PATH) && (p == Src))
                goto do_encode; // First WAR_SLASH in a path. Must be encoded
            goto do_add;
        }
        else if (!isalnum(*p)
            && (*p != '-')
            && (*p != '_')
            && (*p != '\\')
            && (*p != '/')
            && (*p != ':')
            && (*p != '=')
            && (*p != '.'))
        {
do_encode:
            *pp++ = '%';
#if defined(HAVE_ITOA)
            ::itoa((unsigned)((unsigned char)*p++), pp, 16);
#else
            ::sprintf(pp,"%02x", (unsigned)((unsigned char)*p++));
#endif
            pp += 2;
        }
        else
        {
do_add:
            *pp++ = *p++;
        }
    }
    
    *pp = 0;
    Dst = buffer;
    return Dst;
}

==> wfde-3.01.11.tar.gz: wfde-3.01.11/cpplib/WarUrl.cpp <==
#include "StdAfx.h"

#ifndef WAR_ASSERT_H_INCLUDED
#   define WAR_ASSERT_H_INCLUDED
#   include <assert.h>
#endif

#ifndef WAR_MALLOC_H_INCLUDED
#   define WAR_MALLOC_H_INCLUDED
#   include <malloc.h>
#endif

#include "WarUrl.h"   // class implemented

#ifndef WAR_VECTOR_INCLUDED
#   include <vector>
#endif

#ifndef WAR_UTF8_H
#   include "WarUtf8.h"
#endif

using namespace std;


/////////////////////////////// PUBLIC ///////////////////////////////////////

WarUrl& operator << (WarUrl& path, const war_svrpath_t& from)
{
    path.Add(from);
    return path;
}

WarUrl& operator << (WarUrl& path, const WarPathE from)
{
    path.Add(from);
    return path;
}

const char WarUrl::msWarHexToCharMap[127] =
{
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 0
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 10
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 20
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 30
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01, // 40
    0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x00,0x00, // 50
    0x00,0x00,0x00,0x00,0x00,0x0a,0x0b,0x0c,0x0d,0x0e, // 60
    0x0f,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 70
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 80
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x0a,0x0b,0x0c, // 90
    0x0d,0x0e,0x0f,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 100
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 110
    0x00,0x00,0x00,0x00,0x00,0x00,0x00                 // 120
};

//============================= LIFECYCLE ====================================

//============================= OPERATORS ====================================

//============================= OPERATIONS ===================================

void WarUrl::SetHost(war_ccstr_t newHost)
        throw (WarException)
{
    mHostName = "";

    if (!newHost || !*newHost)
        goto rebuild_url;

    if (isalnum(*newHost))
    {
        while ((*newHost != ':') && (*newHost != '/'))
            mHostName += *newHost++;
    }

    if (':' == *newHost)
        mPort = atoi(++newHost);

rebuild_url:
    SetFlagsFromValues();
    BuildUrl();
}

void WarUrl::Create(const WarUrl& defaultUrl, 
                    war_ccstr_t strUrl)
                    throw(WarException)
{
    operator = (defaultUrl);
    ParseString(strUrl, true);
}

void WarUrl::Create(war_ccstr_t type_name,
                    war_ccstr_t UserName,
                    war_ccstr_t Password,
                    war_port_t Port,
                    war_ccstr_t Host,
                    war_ccstr_t UrlPath,
                    war_ccstr_t Command)
                    throw(WarException)
{
    WarUrlType url_type;
    url_type.Create(type_name);
    Create(url_type, 
        UserName,
        Password,
        Port,
        Host,
        UrlPath,
        Command);
}

void WarUrl::Create(const WarUrlType& Type,
                    war_ccstr_t UserName,
                    war_ccstr_t Password,
                    war_port_t Port,
                    war_ccstr_t Host,
                    war_ccstr_t UrlPath,
                    war_ccstr_t Command)
                    throw(WarException)
{
    mUrlType = Type;
    if (UserName && *UserName)
    {
        mUserName = UserName;
        mHaveUserName = true;
    }
    
    if (Password && *Password)
    {
        mPassword = Password;
        mHavePassword = true;
    }
    
    if (Port)
    {
        mPort = Port;
        mHavePort = true;
    }
    
    if (Host && *Host)
    {
        mHostName = Host;
        mHaveHostName = true;
    }
    
    if (UrlPath && *UrlPath)
    {
        mUrlPath = UrlPath;
        mHaveUrlPath = true;
    }
    
    if (Command && *Command)
    {
        mCommand = Command;
        mHaveCommand = true;
    }
    
    BuildUrl();
}

void WarUrl::Reset()
{
    mUrlType.Reset();
    mPort = 0;
    mUserName = "";
    mPassword = "";
    mHostName = "";
    mUrlPath = "";
    mUrl = "";
    mCommand = "";
    mFilePath.Reset();

    mQueryList.erase(mQueryList.begin(), mQueryList.end());
    
    mHavePort = 0;
    mHaveUserName = 0;
    mHavePassword = 0;
    mHaveHostName = 0;
    mHaveUrlPath = 0;
    mHaveCommand = 0;
}

void WarUrl::Normalize()
{
    mFilePath.Normalize();
    mFilePath.RemovePaddingSlash();
    if (mFilePath.mString.find(mFilePath.GetSlash()) == mFilePath.mString.npos)
        mFilePath << WAR_SYSSLASH;
}

void WarUrl::ErasePath()
{
    SetFlagsFromValues();
    mHaveUrlPath = false;
    BuildUrl();
}

//============================= ACCESS     ===================================
//============================= INQUIRY    ===================================


/////////////////////////////// PROTECTED  ///////////////////////////////////

/////////////////////////////// PRIVATE    ///////////////////////////////////


void WarUrl::ParseString(war_ccstr_t src, bool doRelative)
{
    if (doRelative)
    {
        if (!src || !*src)
            return; // Use the current value
    }
    else
        Reset();
    
    size_t url_len = 0;
    war_ccstr_t p = src;
    size_t len = strlen(src) + 8;
    vector<char> tmp_buffer(len);
    war_cstr_t buffer = (war_cstr_t)&tmp_buffer.front();
    war_cstr_t pp = buffer;
    war_ccstr_t pseg1 = NULL, pseg2 = NULL;

    if (doRelative)
    {
        // See if the protocol is identified
        war_ccstr_t p = src;
        if (isalpha(*p))
        {
            while(isalpha(*p))
                ++p;
            if ((*p == ':')
                || ((p[0] == '/')
                && (p[1] == '/')
                && (p[2] == ':')))
            {
                // Looks like a protocol identifier
                // Use default processing.
                ;
            }
            else
            {
consider_relative_path:
                // The protocol identifier is missing, and we
                // are doing relative URL's. This is probarbly
                // path and/or command/query, so we goes directly
                // to that part.
                if (!mUrlType.IsEmpty())
                {
                    SetFlagsFromValues();
                    goto skip_hostname;
                }
            }
        }
        else
            goto consider_relative_path;
    }
    
    try
    {
        WarUrlType url_type;
        url_type.Create(src);
        mUrlType = url_type;
        url_len = mUrlType.GetLenght(); 
    }
    catch(WarException& ex)
    {
        if (!mUrlType.IsEmpty())
        {
            // Use default
            mUrlType.Create(GetTypeName().c_str());
            p += mUrlType.GetLenght();
            goto fetch_path;
        }
        else
        {
            // Assume file if no type is set
#ifdef WIN32
            if (ex == WAR_ERR_INVALID_URL)
            {
                if (!(src 
                    && isalpha(*src)
                    && (src[1] == ':')
                    && ((src[2] == '\\') || (src[2] == '/'))
                    && ((src[3] != '\\') && (src[3] != '/'))))
                    throw ex;    
            }
            else
#endif
            {
                if (ex != WAR_ERR_NOT_AN_URL)
                    throw ex;
            }
            
            mUrlType.Create("file:///");
            goto skip_hostname;
        }
    }
    
    p += url_len;
    
    if (mUrlType.IsHostscanRequiered() && (*p != '/'))
    {
        // We have host, or username, or password, or something...
        // Extract the segment...
        pseg1 = pp = buffer;
        bool bHaveAlpha = false, bHaveCol = false;
        
        while(*p && (*p != '/'))
        {
            switch(*p)
            {
            case ':':
                ++p;
                if (bHaveCol)
                    WarThrow(WarError(WAR_ERR_INVALID_URL), NULL);
                bHaveCol = true;
                *pp++ = 0;
                pseg2 = pp;
                break;
            case '@':
                ++p;
                if (bHaveAlpha)
                    WarThrow(WarError(WAR_ERR_INVALID_URL), NULL);
                bHaveAlpha = true;
                *pp++ = 0;
                mHaveUserName = true;
                if (*pseg1)
                    Decode(mUserName, pseg1);
                if (bHaveCol)
                {
                    // second segment is password
                    mHavePassword = true;
                    assert(pseg2 != NULL);
                    if (*pseg2)
                        Decode(mPassword.mString, pseg2);
                }
                // Reset
                bHaveCol = false;
                pseg1 = pp;
                pseg2 = NULL;
                break;
            default:
                *pp++ = *p++;
                break;
            }
        }
        *pp = 0;
        if (*pseg1)
        {
            mHaveHostName = true;
            Decode(mHostName, pseg1);
        }
        if (bHaveCol)
        {
            mHavePort = true;
            if (!*pseg2)
                WarThrow(WarError(WAR_ERR_INVALID_URL), NULL);
            mPort = atoi(pseg2);
        }
    }

skip_hostname:
    
    if (mUrlType.IsHostscanRequiered() && (*p == '/'))
        ++p;
    
#ifdef WIN32
    else if (!url_len
        && (p[0] == '\\')
        && (p[1] == '\\')
        && (p[2] != '\\')
        && (p[2] != 0))
    {
        // UNC path. Convert to URL
        pseg1 = pp;
        p += 2;
        while(!WarSysSlash::IsSlash(*p))
            *pp++ = *p++;
        
        mHaveHostName = true;
        if (*pseg1)
            mHostName = pseg1;
        
        if (*p)
            ++p;
    }
#endif
    
fetch_path:
    // Until now, we have decoded the URL type, and the user@host part.
    // We now have to decode the url-path and command
    
    if (*p)
    {
        switch(mUrlType.GetStyle())
        {
        case WarUrlType::S_HTTP:
            ParseHttp(p);
            break;
        case WarUrlType::S_FTP:
            ParseFtp(p);
            break;
        case WarUrlType::S_GENERIC:
            ParseGeneric(p);
            break;
        }
    }
    
    BuildUrl();
}

void WarUrl::BuildUrl()
{
    string TmpBuf;
    
    mUrl = GetTypeName();
    
    if (mHaveUserName)
    {
        if (!mUserName.empty())
            mUrl << Encode(TmpBuf, mUserName.c_str());
    }
    
    if (mHavePassword)
    {
        mUrl << ':';
        if (!mPassword.GetValue().empty())
            mUrl << Encode(TmpBuf, mPassword.GetValue().c_str());
    }
    
    if (mHaveUserName || mHavePassword)
        mUrl << '@';
    
    if (mHaveHostName)
    {
        if (!mHostName.empty())
            mUrl << Encode(TmpBuf, mHostName.c_str());
    }
    
    if (mHavePort)
    {
        mUrl << ':';
        WarItoa(mUrl.mString, (unsigned)mPort);
    }
    
    mUrl << '/';
    
    if (mHaveUrlPath && !mUrlPath.empty())
    {
        switch(mUrlType.GetStyle())
        {
        case WarUrlType::S_HTTP:
            {
                war_ccstr_t p = mUrlPath.c_str();
                if (*p == '/')
                    ++p;
                mUrl << Encode(TmpBuf, p, ENC_DEFAULT);
            }
            break;
        default:
            mUrl << Encode(TmpBuf, mUrlPath.c_str(), ENC_PATH);
        }
    }

    switch(mUrlType.GetStyle())
    {
    case WarUrlType::S_HTTP:
        {
            int count = 0;
            for(query_list_t::const_iterator P = mQueryList.begin()
                ; P != mQueryList.end()
                ; ++P)
            {
                if (count++ == 0)
                    mUrl << '?';
                else
                    mUrl << '&';
                string tmp_buf;
                mUrl << Encode(tmp_buf, P->c_str(), ENC_DEFAULT);
            }
        }
        if (mHaveCommand)
        {
            string tmp_buf;
            mUrl << '#' << Encode(tmp_buf, mCommand.c_str(), ENC_DEFAULT);;
        }
        break;
    case WarUrlType::S_FTP:
        if (mHaveCommand)
        {
            string tmp_buf;
            mUrl << ';' << Encode(tmp_buf, mCommand.c_str(), ENC_DEFAULT);;
        }
        break;
    case WarUrlType::S_GENERIC:
        break;
    }
    
    if (mUrlType == "file")
    {
        // Make a system file path!
        mFilePath.Reset();
        if (mHaveHostName && !mHostName.empty())
        {
            mFilePath << WAR_DSYSSLASH
                << WarUtf8(mHostName).GetUnicode()
                << WAR_SYSSLASH;
        }
        
        if (mHaveUrlPath && !mUrlPath.empty())
        {
            war_ccstr_t my_path = mUrlPath.c_str();
            bool start_is_slash = mFilePath.IsSlash(*my_path);
            if (start_is_slash 
                && mFilePath.IsEmpty() 
                && mFilePath.IsSlash(mFilePath.GetLastCh()))
                ++my_path;
            else if (!start_is_slash)
            {
#ifdef WIN32
                // Avoid ./ prefix on C:\  ...
                if (!(isalpha(my_path[0]) && (':' == my_path[1])))
#endif
                {
                    mFilePath << WAR_CDIR;
                }
            }
#ifdef WIN32
            // Windows don't like paths like "\\server\\share\blah",
            // so we have to strip off the leading path if it
            // exist...
            if (mHaveHostName 
                && *my_path 
                && mFilePath.IsSlash(*my_path))
                ++my_path;
#endif
            if (*my_path)
            {
                mFilePath << WarUtf8(my_path).GetUnicode();
            }
        }
        else
            mFilePath << WAR_CDIR;
        mFilePath.Deparse();
        mFilePath.Normalize();
    }

    if (!mHavePassword || GetPassword().empty())
    {
        mSafeUrl = mUrl.GetValue();
    }
    else
    {
        // Hide the password
        war_ccstr_t p = mUrl.GetValue().c_str();
        char *buf = (char *)alloca(mUrl.GetValue().size() + 8);
        char *pp = buf;

        while(*p && (*p != ':'))
            *pp++ = *p++;
        if (*p == ':')
            *pp++ = *p++;
        if (*p == '/')
            *pp++ = *p++;
        if (*p == '/')
            *pp++ = *p++;
        while(*p && (*p != ':'))
            *pp++ = *p++;
        if (*p == ':')
            *pp++ = *p++;
        while(*p && (*p != '@'))
            +p++;
        *pp++ = '*';
        *pp++ = '*';
        *pp++ = '*';
        *pp++ = '*';
        while(*p)
            *pp++ = *p++;
        *pp = 0;
        mSafeUrl = buf;
    }
}

std::string& WarUrl::Encode(std::string& Dst, 
                            war_ccstr_t Src, 
                            const EncodeModeE encMode)
{
    war_ccstr_t p = Src;
    size_t max_len = (strlen(Src) * 3)+ 1;
    
    war_cstr_t buffer = (war_cstr_t)alloca(max_len);
    war_cstr_t pp = buffer;
    
    while(*p)
    {
        if (*p == '/')
        {
            if ((encMode == ENC_PATH) && (p == Src))
                goto do_encode; // First WAR_SLASH in a path. Must be encoded
            goto do_add;
        }
        else if (!isalnum(*p)
            && (*p != '-')
            && (*p != '_')
            && (*p != '\\')
            && (*p != '/')
            && (*p != ':')
            && (*p != '=')
            && (*p != '.'))
        {
do_encode:
            *pp++ = '%';
#if defined(HAVE_ITOA)
            ::itoa((unsigned)((unsigned char)*p++), pp, 16);
#else
            ::sprintf(pp,"%02x", (unsigned)((unsigned char)*p++));
#endif
            pp += 2;
        }
        else
        {
do_add:
            *pp++ = *p++;
        }
    }
    
    *pp = 0;
    Dst = buffer;
    return Dst;
}

void WarUrl::ParseHttp(war_ccstr_t str)
{
    war_ccstr_t p = str;
    
    while(*p && (*p != '?') && (*p != '#'))
        ++p;

    if (*p == 0)
    {
        // Clean path
        Decode(mUrlPath, str);
        mHaveUrlPath = true;
        return;
    }

    size_t path_len = p - str;
    string buffer;
    buffer.resize(path_len);
    memcpy(&buffer[0], str, path_len);
    Decode(mUrlPath, str);
    mHaveUrlPath = true;

    if ((*p == '?') && *++p)
    {
        do
        {
            war_ccstr_t start_ofs = p;
            string decoded_str;
            while(*p && (*p != '&') && (*p != '#'))
            {
                path_len = p - start_ofs;
                buffer.resize(path_len);
                memcpy(&buffer[0], start_ofs, path_len);
                decoded_str = "";
                Decode(decoded_str, buffer.c_str());
                mQueryList.push_back(decoded_str);
            }
        } while(*p && (*p == '&'));
    }

    if ((*p == '#') && *++p)
    {
        mHaveCommand = true;
        Decode(mCommand, p); 
    }
}
    
void WarUrl::ParseFtp(war_ccstr_t str)
{
    vector<char> buf(strlen(str) +1);
    char *p = &buf[0];
    while(*str && (';' != *str))
        *p++ = *str++;
    *p = 0;
    Decode(mUrlPath, str);
    mHaveUrlPath = !mUrlPath.empty();

    if (*str && *++str)
    {
        Decode(mCommand, str);
        mHaveCommand = true;
    }
}

void WarUrl::ParseGeneric(war_ccstr_t str)
{
    if (*str)
    {
        Decode(mUrlPath, str);
        mHaveUrlPath = true;
    }
}

void WarUrl::SetFlagsFromValues()
{
    if (mPort)
        mHavePort = true;
    if (!mUserName.empty())
        mHaveUserName = true;
    if (!mPassword.GetValue().empty())
        mHavePassword = true;
    if (!mUrlPath.empty())
        mHaveUrlPath = true;
    if (!mCommand.empty())
        mHaveCommand = true;
    if (!mHostName.empty())
        mHaveHostName = true;
}
